from importlib import import_module
from warnings import warn

_deprecations = {"AppContext": "delfino.models", "pass_app_context": "delfino.decorators"}


def __getattr__(name):
    if name in _deprecations:
        new_module = _deprecations[name]
        warn(
            f"Import of `{name}` from `{__name__}` is deprecated. Use `{new_module}` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return getattr(import_module(new_module), name)
    raise AttributeError(f"module {__name__} has no attribute {name}")



from logging import getLogger
from typing import Any, Dict, Optional
from warnings import warn

import click
from click import BadParameter

from delfino.models import AppContext

_LOG = getLogger(__name__)


class SetOptionFromConfigCallback:
    """If config option exists in config and not set via command line, it will be used instead.

    Config option is looked up at ``tool.delfino.plugins.<PLUGIN>.<COMMAND NAME>.<COMMAND_ARGUMENT_NAME>``.
    """

    def __init__(self, command_argument_name: str, config_option_name: Optional[str] = None):
        self.config_option_name = config_option_name or command_argument_name
        self.command_argument_name = command_argument_name

    def _type_cast_value(self, ctx: click.Context, param: click.Parameter, value_from_config: Any) -> Any:
        try:
            return param.type_cast_value(ctx, value_from_config)
        except BadParameter as exc:
            exc.param_hint = f"the '{ctx.info_name}.{self.config_option_name}' config option in pyproject.toml file"
            raise

    def set_parameter_from_config_in_group(self, ctx: click.Context, command: click.Command):
        """Same as ``__call__`` but called on a given command.

        This is useful for invoking commands indirectly, for example in a group command.
        """
        warn(
            "This method is deprecated. Use `parameter_from_config_in_group` instead.", DeprecationWarning, stacklevel=1
        )
        app_context = ctx.find_object(AppContext)

        if app_context is None:
            raise RuntimeError("AppContext was expected to be set but none found.")

        command_config = getattr(app_context.plugin_config, command.name or "", {})

        if isinstance(command_config, dict):
            value_from_config = command_config.get(self.config_option_name, None)
        else:
            value_from_config = getattr(command_config, self.config_option_name, None)

        if value_from_config:
            for command_param in command.params:
                if isinstance(command_param, click.Parameter) and command_param.name == self.command_argument_name:
                    ctx.params[self.command_argument_name] = self._type_cast_value(
                        ctx, command_param, value_from_config
                    )
                    return
            _LOG.warning(
                f"Command '{command.name}' has no parameter '{self.command_argument_name}'. Please update the "
                f"pyproject.toml file, option '{command.name}.{self.config_option_name}'."
            )

    def parameter_from_config_in_group(self, ctx: click.Context, command: click.Command) -> Dict[str, Any]:
        """Returns key-value pair to set in invoking another command.

        Similiarly to ``__call__``, it takes the value from config if it exists. It also checks
        if the command has a ``click.Parameter`` of the matching name.

        This is useful for invoking commands indirectly, for example in a group command.
        """
        app_context = ctx.find_object(AppContext)

        if app_context is None:
            raise RuntimeError("AppContext was expected to be set but none found.")

        command_config = getattr(app_context.plugin_config, command.name or "", {})

        if isinstance(command_config, dict):
            value_from_config = command_config.get(self.config_option_name, None)
        else:
            value_from_config = getattr(command_config, self.config_option_name, None)

        if value_from_config:
            for command_param in command.params:
                if isinstance(command_param, click.Parameter) and command_param.name == self.command_argument_name:
                    return {self.command_argument_name: self._type_cast_value(ctx, command_param, value_from_config)}
            _LOG.warning(
                f"Command '{command.name}' has no parameter '{self.command_argument_name}'. Please update the "
                f"pyproject.toml file, option '{command.name}.{self.config_option_name}'."
            )
        return {}

    def __call__(self, ctx: click.Context, param: click.Parameter, value: Any) -> Any:
        """Load passed arguments from config if none given on command line."""
        # Command line options have higher priority. Return them instead of options from the config file.
        if value:
            return value

        # No command line options provided, try to load them from config
        app_context = ctx.find_object(AppContext)

        if app_context is None:
            raise RuntimeError("AppContext was expected to be set but none found.")

        command_name: str = ctx.command.name or ""
        value_from_config: str = getattr(app_context.plugin_config, command_name, {}).get(self.config_option_name, None)
        if value_from_config:
            return self._type_cast_value(ctx, param, value_from_config)

        return value



from pathlib import Path
from typing import Generic, TypeVar

from pydantic import BaseModel

from delfino.constants import PackageManager
from delfino.models.pyproject_toml import PluginConfig, PyprojectToml

PluginConfigType = TypeVar("PluginConfigType", bound=PluginConfig)


class AppContext(Generic[PluginConfigType], BaseModel):
    project_root: Path
    pyproject_toml: PyprojectToml
    package_manager: PackageManager
    plugin_config: PluginConfigType

    class Config:
        arbitrary_types_allowed = True



import logging
from typing import Union

import click

_NAME_TO_LEVEL = {
    "CRITICAL": logging.CRITICAL,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
}


def _set_log_level(ctx: click.Context, param: Union[click.Option, click.Parameter], value: str):
    del param

    if ctx.resilient_parsing:
        return

    log_level = _NAME_TO_LEVEL.get(value.upper(), logging.INFO)

    logging.basicConfig(level=log_level)
    for logger_name in logging.root.manager.loggerDict:  # pylint: disable=no-member
        logging.getLogger(logger_name).setLevel(log_level)


log_level_option = click.option(
    "--log-level",
    type=click.Choice(list(_NAME_TO_LEVEL.keys()), case_sensitive=False),
    default="INFO",
    is_eager=True,
    show_default=True,
    callback=_set_log_level,
    help="Log level of the output.",
)



from typing import Union

import click


def _print_help(ctx: click.Context, param: Union[click.Option, click.Parameter], value: bool):
    del param
    if not value or ctx.resilient_parsing:
        return
    click.echo(ctx.get_help())
    ctx.exit()


extended_help_option = click.option(
    "-h",
    "--help",
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=_print_help,
    help="Show this message and exit.",
)
"""Adds ``-h`` alongside the built-in ``--help`` option."""



import click

from delfino.click_utils.set_from_config import SetOptionFromConfigCallback

_ARGUMENT_NAME: str = "files_folders"
FILES_FOLDERS_OPTION_CALLBACK = SetOptionFromConfigCallback(_ARGUMENT_NAME)

files_folders_option = click.option(
    "-f",
    "--file",
    "--folder",
    _ARGUMENT_NAME,
    multiple=True,
    nargs=1,
    type=click.Path(exists=True),
    help="A file or a folder to pass to the downstream tool instead of the default ones. "
    "Can be supplied multiple times.",
    callback=FILES_FOLDERS_OPTION_CALLBACK,
)



import shlex
from typing import Any

import click

from delfino.click_utils.set_from_config import SetOptionFromConfigCallback


class _SetPassArgsFromConfigCallback(SetOptionFromConfigCallback):
    def _type_cast_value(self, ctx: click.Context, param: click.Parameter, value_from_config: Any) -> Any:
        if isinstance(value_from_config, str):
            value_from_config = tuple(shlex.split(value_from_config))

        return super()._type_cast_value(ctx, param, value_from_config)


_ARGUMENT_NAME = "passed_args"
PASS_ARGS_CALLBACK = _SetPassArgsFromConfigCallback(_ARGUMENT_NAME, "pass_args")


pass_args = click.argument(
    _ARGUMENT_NAME,
    metavar=f"[-- {_ARGUMENT_NAME.upper()}]",
    nargs=-1,
    callback=PASS_ARGS_CALLBACK,
    type=click.UNPROCESSED,
)
"""A command decorator which passes additional arguments to the decorated command.

Example:

    @click.command("test")
    @pass_args
    def run_pytest(passed_args: List[str]):
        run(["pytest", *passed_args])

This decorator adds a ``passed_args`` parameter containing all the additional passed arguments
from one of the following sources:
    - The ``pyproject.toml`` file, under ``tools.delfino.<PLUGIN>.<COMMAND>.pass_args``. Example:

        [tool.delfino.plugins.delfino-core.test]
        pass_args = '--capture=no'

    - From commands line, capturing anything after ``--``. Example:

        delfino test -- --capture=no

Note:
    This decorator will not automatically pass the argument to the wrapped
    command. You need to do it manually.
"""



import functools
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, List, Union

import click

from delfino.constants import ENTRY_POINT
from delfino.validation import assert_pip_package_installed

try:
    import shellingham
except ImportError:
    pass


class CompletionAlreadyInstalled(Exception):
    pass


@dataclass
class Completion:
    shell: str
    completion: str
    install: Callable[[str], Path]

    @property
    def formatted_completion(self) -> str:
        return self.completion.format(entry_point=ENTRY_POINT, entry_point_upper=ENTRY_POINT.upper()).strip()


_COMPLETION_BASH = """
_complete_{entry_point}() {{
    [ -n "$(which {entry_point})" ] && \\
        eval "$(_{entry_point_upper}_COMPLETE=bash_source {entry_point})";
}}
complete -F _complete_{entry_point} -o default invoke {entry_point}
"""

_COMPLETION_ZSH = """
_complete_{entry_point}() {{
    which {entry_point} >/dev/null && \\
        eval $(_{entry_point_upper}_COMPLETE=zsh_source {entry_point})
}}
compctl -K _complete_{entry_point} + -f invoke {entry_point}
"""


def handle_assertion_error(func):
    @functools.wraps(func)
    def inner(ctx: click.Context, param: Union[click.Option, click.Parameter], value: bool):
        try:
            return func(ctx, param, value)
        except AssertionError as exc:
            click.secho(str(exc), fg="red", err=True)
            raise click.exceptions.Exit(1) from exc

    return inner


def _install_completion(
    formatted_completion: str, *, completion_path: Path, completion_init_lines: List[str], rc_path: Path
) -> Path:
    rc_path.parent.mkdir(parents=True, exist_ok=True)
    rc_content = rc_path.read_text(encoding="utf-8") if rc_path.is_file() else ""

    content_added = False
    for line in completion_init_lines:
        if line not in rc_content:  # pragma: nocover
            rc_content += f"\n{line}"
            content_added = True

    if not content_added:
        raise CompletionAlreadyInstalled()

    rc_content += "\n"
    rc_path.write_text(rc_content)

    # Install completion
    completion_path.parent.mkdir(parents=True, exist_ok=True)
    completion_path.write_text(formatted_completion)
    return completion_path


def _install_completion_bash(formatted_completion: str) -> Path:
    completion_path = Path.home() / f".bash_completions/{ENTRY_POINT}.sh"
    return _install_completion(
        formatted_completion,
        completion_path=completion_path,
        completion_init_lines=[f"source {completion_path}"],
        rc_path=Path.home() / ".bashrc",
    )


def _install_completion_zsh(formatted_completion: str) -> Path:
    return _install_completion(
        formatted_completion,
        completion_path=Path.home() / f".zfunc/_{ENTRY_POINT}",
        completion_init_lines=[
            "autoload -Uz compinit",
            "compinit",
            "zstyle ':completion:*' menu select",
            "fpath+=~/.zfunc",
        ],
        rc_path=Path.home() / ".zshrc",
    )


_COMPLETIONS: Dict[str, Completion] = {
    "bash": Completion(shell="bash", completion=_COMPLETION_BASH, install=_install_completion_bash),
    "zsh": Completion(shell="zsh", completion=_COMPLETION_ZSH, install=_install_completion_zsh),
}


def _get_completion_for_current_shell(param: Union[click.Option, click.Parameter]) -> Completion:
    assert_pip_package_installed("shellingham", required_by=f"{param.param_type_name} --{param.name}")

    shell: str = shellingham.detect_shell()[0]
    try:
        return _COMPLETIONS[shell]
    except KeyError as exc:
        supported_shells = ", ".join(_.capitalize() for _ in _COMPLETIONS)
        click.secho(f"{shell.capitalize()} is not supported. Supported shells are: {supported_shells}", fg="red")
        raise click.Abort() from exc


@handle_assertion_error
def _show_completion_for_current_shell(ctx: click.Context, param: Union[click.Option, click.Parameter], value: bool):
    if not value or ctx.resilient_parsing:
        return
    click.secho(_get_completion_for_current_shell(param).formatted_completion)
    raise click.exceptions.Exit(0)


show_completion_option = click.option(
    "--show-completion",
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=_show_completion_for_current_shell,
    help="Show completion for the current shell, to copy it or customize the installation.",
)


def _install_completion_for_current_shell(ctx: click.Context, param: Union[click.Option, click.Parameter], value: bool):
    if not value or ctx.resilient_parsing:
        return
    completion = _get_completion_for_current_shell(param)
    try:
        path = completion.install(completion.formatted_completion)
        click.secho(f"✔ {completion.shell.capitalize()} completion installed in '{path}'.", fg="green")
        click.secho("⚠ Completion will take effect once you restart the terminal.", fg="yellow")
    except CompletionAlreadyInstalled:
        click.secho(
            f"✔ {completion.shell.capitalize()} completion has been previously installed. Not installing again.",
            fg="green",
        )

    raise click.exceptions.Exit(0)


install_completion_option = click.option(
    "--install-completion",
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=_install_completion_for_current_shell,
    help="Install completion for the current shell.",
)



import logging
import os
import sys
from pathlib import Path
from typing import Any, List, Optional

import click
import toml
from pydantic import ValidationError

from delfino.click_utils.command import CommandRegistry
from delfino.click_utils.completion import install_completion_option, show_completion_option
from delfino.click_utils.help import extended_help_option
from delfino.click_utils.verbosity import log_level_option
from delfino.constants import ENTRY_POINT, PYPROJECT_TOML_FILENAME
from delfino.models.app_context import AppContext
from delfino.models.pyproject_toml import PyprojectToml
from delfino.utils import get_package_manager


class Commands(click.MultiCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        sys.path.append(os.getcwd())

        self._project_root = Path(os.getcwd())
        pyproject_toml_path = (self._project_root / PYPROJECT_TOML_FILENAME).relative_to(self._project_root)

        # When evaluating available commands, the program should not fail. We save the exception
        # to show it only when a command is executed.
        self._pyproject_toml_validation_error: Optional[ValidationError] = None

        try:
            self._pyproject_toml = PyprojectToml(file_path=pyproject_toml_path, **toml.load(pyproject_toml_path))
        except ValidationError as exc:
            self._pyproject_toml = PyprojectToml()
            self._pyproject_toml_validation_error = exc
        except FileNotFoundError:
            self._pyproject_toml = PyprojectToml()

        self._command_registry = CommandRegistry(
            plugins_configs=self._pyproject_toml.tool.delfino.plugins,
            local_commands_directory=self._pyproject_toml.tool.delfino.local_commands_directory,
        )

    def list_commands(self, ctx: click.Context) -> List[str]:
        """Override as MultiCommand always returns []."""
        del ctx
        return sorted(command.name for command in self._command_registry.visible_commands)

    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        """Override to give all commands a common ``AppContext`` or fail if ``pyproject.toml`` is broken/missing."""
        cmd = self._command_registry.get(cmd_name, None)
        if cmd is None:
            return None  # command doesn't exist

        if ctx.resilient_parsing:  # do not fail on auto-completion
            return cmd.command

        if self._pyproject_toml_validation_error:
            click.secho(
                f"Delfino appears to be misconfigured: {self._pyproject_toml_validation_error}", fg="red", err=True
            )
            raise click.Abort() from self._pyproject_toml_validation_error

        ctx.obj = AppContext(
            project_root=self._project_root,
            pyproject_toml=self._pyproject_toml,
            package_manager=get_package_manager(self._project_root, self._pyproject_toml),
            plugin_config=cmd.package.plugin_config,
        )

        return cmd.command

    def invoke(self, ctx: click.Context) -> Any:
        """Override to turn ``AssertionError`` exception into ``click.exceptions.Exit``."""
        try:
            return super().invoke(ctx)
        except AssertionError as exc:
            click.secho(f"Command '{ctx.invoked_subcommand}' is misconfigured. {exc}", fg="red", err=True)
            raise click.exceptions.Exit(1)

    def get_help(self, *args, **kwargs) -> str:
        help_str = super().get_help(*args, **kwargs)

        if self._command_registry.hidden_commands:
            if logging.root.level == logging.DEBUG:
                help_str += click.style(
                    f"\n\nDisabled command{'s' if len(self._command_registry.hidden_commands) > 1 else ''}: "
                    + ", ".join(
                        sorted(
                            [
                                f"{command.package.plugin_name}/{command.command.name}"
                                for command in self._command_registry.hidden_commands
                            ]
                        )
                    )
                    + f" (see 'tool.{ENTRY_POINT}.<PLUGIN NAME>.disable_commands' in '{PYPROJECT_TOML_FILENAME}')",
                    fg="white",
                )

        return help_str


@click.group(cls=Commands)
@extended_help_option
@click.version_option()
@log_level_option
@show_completion_option
@install_completion_option
def main(log_level=None):
    del log_level


if __name__ == "__main__":
    main()



from enum import Enum
from pathlib import Path

ENTRY_POINT: str = "delfino"
PYPROJECT_TOML_FILENAME = "pyproject.toml"
DEFAULT_LOCAL_COMMANDS_DIRECTORY = Path("commands")


class PackageManager(Enum):
    POETRY = "poetry"
    PIPENV = "pipenv"
    UNKNOWN = "unknown"



import sys
from subprocess import PIPE

from delfino.constants import PYPROJECT_TOML_FILENAME, PackageManager
from delfino.execution import OnError, run

if sys.version_info <= (3, 8):
    import importlib_metadata as metadata
else:
    from importlib import metadata


def pyproject_toml_key_missing(key: str) -> str:
    return f"Key '{key}' is missing from the '{PYPROJECT_TOML_FILENAME}' file."


def pip_package_installed(name: str, sub_process: bool = False) -> bool:
    if sub_process:
        return run(f"pip show -q {name}", stdout=PIPE, stderr=PIPE, on_error=OnError.PASS, shell=True).returncode == 0

    try:
        metadata.Distribution.from_name(name)
    except metadata.PackageNotFoundError:
        return False

    return True


def assert_pip_package_installed(name: str, required_by: str = "this command") -> None:
    assert pip_package_installed(
        name
    ), f"Optional Python package '{name}' is required by {required_by} but not installed."


def assert_package_manager_is_known(package_manager: PackageManager) -> None:
    assert (
        package_manager != PackageManager.UNKNOWN
    ), "Cannot determine package manager used in this project. Only the following ones are supported: " + ", ".join(
        member.value for member in PackageManager.__members__.values() if member != PackageManager.UNKNOWN
    )



import functools
from typing import Any, Callable, Type, TypeVar, cast

from click import get_current_context

from delfino.models.app_context import AppContext
from delfino.models.pyproject_toml import PluginConfig

_Func = TypeVar("_Func", bound=Callable[..., Any])


def pass_app_context(
    plugin_config_type: Type[PluginConfig] = PluginConfig, kwargs_name: str = "app_context"
) -> Callable[[_Func], _Func]:
    """Similar to ``click.make_pass_decorator``, with optional parsing of plugin specific config.

    This decorator passes into the command an app context ``AppContext``. In this context,
    the ``plugin_config`` attribute holds a config specific to the plugin where the called
    command originates. Any unknown fields will be kept. If you want to parse and validate
    this config, pass a subclass of ``PluginConfig`` as the ``plugin_config_type`` argument
    of this decorator. ``plugin_config`` will be then replaced with this instance, provided
    it is valid.

    Args:
        plugin_config_type: A subclass of ``PluginConfig``.
        kwargs_name: Name of the argument passed to the decorated command.
    """

    def decorator(func: _Func) -> _Func:
        def new_func(*args, **kwargs):
            ctx = get_current_context()
            obj = ctx.find_object(AppContext)
            if obj is None:
                raise RuntimeError(
                    f"Managed to invoke callback without a context object of type {AppContext.__name__!r} existing."
                )
            obj.plugin_config = plugin_config_type(**obj.plugin_config.dict())
            return ctx.invoke(func, *args, **kwargs, **{kwargs_name: obj})

        return functools.update_wrapper(cast(_Func, new_func), func)

    return decorator



from pathlib import Path
from typing import Any, Dict, Optional, Set

from pydantic import BaseModel, Extra, Field

from delfino.constants import DEFAULT_LOCAL_COMMANDS_DIRECTORY


class PluginConfig(BaseModel):
    enable_commands: Set[str] = set()
    disable_commands: Set[str] = set()

    @classmethod
    def empty(cls):
        return cls()

    class Config:
        extra = Extra.allow  # Allows arbitrary plugin-specific keys


class Delfino(BaseModel):
    local_commands_directory: Path = DEFAULT_LOCAL_COMMANDS_DIRECTORY
    plugins: Dict[str, PluginConfig] = Field(default_factory=dict)

    class Config:
        extra = Extra.allow


class Poetry(BaseModel):
    name: str
    version: str
    scripts: Dict[str, str] = Field(default_factory=dict)
    dependencies: Dict[str, Any] = Field(default_factory=dict)

    class Config:
        extra = Extra.allow


class Tool(BaseModel):
    poetry: Optional[Poetry] = None
    delfino: Delfino = Field(default_factory=Delfino)

    class Config:
        allow_population_by_field_name = True


class PyprojectToml(BaseModel):
    file_path: Optional[Path] = None
    tool: Tool = Field(default_factory=Tool)

    class Config:
        extra = Extra.allow



from pathlib import Path
from typing import List, Union

from delfino.constants import PackageManager
from delfino.models.pyproject_toml import PyprojectToml

ArgsList = List[Union[str, bytes, Path]]
ArgsType = Union[str, bytes, List]


def get_package_manager(project_root: Path, pyproject_toml: PyprojectToml) -> PackageManager:
    if pyproject_toml.tool.poetry is not None or (project_root / "poetry.lock").exists():
        return PackageManager.POETRY
    if (project_root / "Pipfile").exists() or (project_root / "Pipfile.lock").exists():
        return PackageManager.PIPENV

    return PackageManager.UNKNOWN



import shlex
from typing import Iterable


def shlex_join(split_command: Iterable[str]) -> str:
    """Return a shell-escaped string from *split_command*.

    Backport of ``shlex.join`` for Python 3.7
    """
    return " ".join(shlex.quote(arg) for arg in split_command)



import logging
import sys
from dataclasses import dataclass
from importlib import import_module, resources
from importlib.resources import Package
from pathlib import Path
from typing import Dict, Iterator, List, Mapping, Optional, cast

import click
from pydantic import BaseModel, Field, validator

from delfino.constants import DEFAULT_LOCAL_COMMANDS_DIRECTORY
from delfino.models.pyproject_toml import PluginConfig

if sys.version_info < (3, 10):
    # `importlib.metadata` available since 3.8 but `distribution.entry_points.select` only since 3.10
    from importlib_metadata import distributions
else:
    from importlib.metadata import distributions


_LOG = logging.getLogger(__name__)


def command_names(commands: List[click.Command]) -> str:
    return ", ".join(command.name for command in commands if command.name)


class _CommandPackage(BaseModel):
    """Holds `importlib.resources.Package` that include commands and metadata for delfino."""

    plugin_name: str = Field(..., description="Name of the plugin.")
    package: Package = Field(..., description="A package that include commands.")
    plugin_config: PluginConfig

    @validator("package")
    def package_is_valid(cls, value: Package):  # pylint: disable=no-self-argument
        if isinstance(value, str):
            return value
        if isinstance(value.__package__, str) and isinstance(value.__file__, str):
            return value

        raise ValueError(
            f"`package` must be either an instance of `str` or have the `__module__` and "
            f"`__file__` attributes set to a `str` but '{value}' used."
        )

    @property
    def module_name(self) -> str:
        if isinstance(self.package, str):
            return self.package
        assert isinstance(self.package.__package__, str)
        return self.package.__package__

    @property
    def module_root_dir(self) -> Path:
        if isinstance(self.package, str):
            return Path(self.package)
        assert isinstance(self.package.__file__, str)
        return Path(self.package.__file__).parent

    class Config:
        arbitrary_types_allowed = True


@dataclass(frozen=True)
class _Command:
    name: str
    command: click.Command
    package: _CommandPackage


def find_commands(command_package: _CommandPackage) -> List[_Command]:
    """Finds all instances of ``click.Command`` in given module.

    Does not traverse modules recursively. Plugins must point to the correct module via the entry point.
    """
    try:
        if sys.version_info >= (3, 9):
            files = [
                path.name for path in resources.files(command_package.package).iterdir()  # pylint: disable=no-member
            ]
        else:
            files = resources.contents(command_package.package)
    except ModuleNotFoundError:
        return []

    commands: List[_Command] = []

    for filename in files:
        if not filename.endswith(".py") or (filename.startswith("_") and filename != "__init__.py"):
            continue
        module = import_module(f"{command_package.module_name}.{filename[:-3]}")

        commands.extend(
            _Command(name=obj.name, command=obj, package=command_package)
            for obj_name, obj in vars(module).items()
            if not obj_name.startswith("_") and isinstance(obj, click.Command) and obj.name is not None
        )

    return commands


def get_root_command(click_context: click.Context) -> click.MultiCommand:
    """Find the root command.

    In the context of `delfino`, this is generally the ``main.Commands`` instance.
    """
    while click_context.parent:
        click_context = click_context.parent
    return cast(click.MultiCommand, click_context.command)


class CommandRegistry(Mapping):
    """An internal command registry which can raise error with duplicated commands.

    The purpose of this class is to keep track on registered commands and raise an
    error if a command with the same name is registered twice.

    The error will be raised only when a command is from plugins. Local commands can
    always override existing commands.
    """

    TYPE_OF_PLUGIN = "delfino.plugin"
    LOCAL_PLUGIN_NAME = "local"

    def __init__(
        self,
        plugins_configs: Dict[str, PluginConfig],
        command_packages: Optional[List[_CommandPackage]] = None,
        local_commands_directory: Path = DEFAULT_LOCAL_COMMANDS_DIRECTORY,
    ):
        if command_packages is None:
            self._command_packages = self._default_command_packages(plugins_configs, local_commands_directory)
        else:
            self._command_packages = command_packages
        self._visible_commands: Dict[str, _Command] = {}
        self._hidden_commands: Dict[str, _Command] = {}
        self._register_packages()

    def __len__(self) -> int:
        return len(self._visible_commands)

    def __getitem__(self, key: str) -> _Command:
        return self._visible_commands[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._visible_commands)

    @property
    def visible_commands(self) -> List[_Command]:
        return list(self._visible_commands.values())

    @property
    def hidden_commands(self) -> List[_Command]:
        return list(self._hidden_commands.values())

    @classmethod
    def _default_command_packages(
        cls, plugins_configs: Dict[str, PluginConfig], local_commands_directory: Path
    ) -> List[_CommandPackage]:
        # This is a function to lazy load packages in tests. They may not exist on import of the code.
        return [
            # Lower priority - discovered installed packages
            *cls._discover_command_packages(plugins_configs),
            # Higher priority - locally available packages
            _CommandPackage(
                plugin_name=cls.LOCAL_PLUGIN_NAME,
                package=str(local_commands_directory),
                plugin_config=plugins_configs.get(cls.LOCAL_PLUGIN_NAME, PluginConfig.empty()),
            ),
        ]

    @classmethod
    def _discover_command_packages(cls, plugins_configs: Dict[str, PluginConfig]) -> List[_CommandPackage]:
        """Discover packages from plugin. It is using package metadata as plugin discovering solution.

        Check the following URL about the plugin discovering solutions including the one uses package metadata.
        https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/
        """
        # We want to keep loaded _CommandPackage instance in the same order as defined in the config
        command_packages: Dict[str, Optional[_CommandPackage]] = {plugin_name: None for plugin_name in plugins_configs}
        for distribution in distributions():
            for entry_point in distribution.entry_points.select(group=cls.TYPE_OF_PLUGIN):
                if not entry_point:
                    continue
                plugin_name = distribution.metadata["Name"]
                if plugin_name not in command_packages:
                    continue
                command_packages[plugin_name] = _CommandPackage(
                    plugin_name=plugin_name,
                    package=entry_point.load(),
                    plugin_config=plugins_configs.get(plugin_name, PluginConfig.empty()),
                )

        found_command_packages = []
        for plugin_name, command_package in list(command_packages.items()):
            if command_package is not None:
                found_command_packages.append(command_package)
            else:
                if plugin_name != cls.LOCAL_PLUGIN_NAME:
                    _LOG.warning(f"Plugin '{plugin_name}' specified in config but no such plugin is installed.")
                command_packages.pop(plugin_name)

        return found_command_packages

    def _register_packages(self):
        for command_package in self._command_packages:
            commands = {command.name: command for command in find_commands(command_package)}
            enabled_commands = command_package.plugin_config.enable_commands or set(commands.keys())
            enabled_commands.difference_update(command_package.plugin_config.disable_commands)

            for command_name, command in commands.items():
                self._register(command, command_name in enabled_commands)

    def _register(self, command: _Command, enabled: bool):
        existing_command = self._visible_commands.pop(command.name, None) or self._hidden_commands.pop(
            command.name, None
        )
        if existing_command:
            _LOG.debug(
                f"Using command '{command.name}' from plugin '{command.package.plugin_name}'. Previously registered "
                f"by '{existing_command.package.plugin_name}' plugin, which has lower priority."
            )

        if enabled:
            self._visible_commands[command.name] = command
        else:
            self._hidden_commands[command.name] = command
            _LOG.debug(
                f"Command '{command.name}' from the '{command.package.plugin_name}' plugin has been disabled in config."
            )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['delfino', 'delfino.click_utils', 'delfino.decorators', 'delfino.models']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0,<9.0',
 'deprecation>=2.1,<3.0',
 'pydantic>=1.8,<2.0',
 'toml>=0.10,<0.11']

extras_require = \
{':python_version < "3.10"': ['importlib-metadata']}

entry_points = \
{'console_scripts': ['delfino = delfino.main:main', 'mike = delfino.main:main']}

setup_kwargs = {
    'name': 'delfino',
    'version': '0.29.0',
    'description': 'A collection of command line helper scripts wrapping tools used during Python development.',
    'long_description': '<h1 align="center" style="border-bottom: none;">🧰&nbsp;&nbsp;Delfino&nbsp;&nbsp;🧰</h1>\n<h3 align="center">A collection of command line helper scripts wrapping tools used during Python development.</h3>\n\n<p align="center">\n    <a href="https://app.circleci.com/pipelines/github/radeklat/delfino?branch=main">\n        <img alt="CircleCI" src="https://img.shields.io/circleci/build/github/radeklat/delfino">\n    </a>\n    <a href="https://app.codecov.io/gh/radeklat/delfino/">\n        <img alt="Codecov" src="https://img.shields.io/codecov/c/github/radeklat/delfino">\n    </a>\n    <a href="https://github.com/radeklat/delfino/tags">\n        <img alt="GitHub tag (latest SemVer)" src="https://img.shields.io/github/tag/radeklat/delfino">\n    </a>\n    <img alt="Maintenance" src="https://img.shields.io/maintenance/yes/2022">\n    <a href="https://github.com/radeklat/delfino/commits/main">\n        <img alt="GitHub last commit" src="https://img.shields.io/github/last-commit/radeklat/delfino">\n    </a>\n    <a href="https://www.python.org/doc/versions/">\n        <img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/delfino">\n    </a>\n    <a href="https://pypistats.org/packages/delfino">\n        <img alt="Downloads" src="https://img.shields.io/pypi/dm/delfino">\n    </a>\n</p>\n\n<!--\n    How to generate TOC from PyCharm:\n    https://github.com/vsch/idea-multimarkdown/wiki/Table-of-Contents-Extension\n-->\n[TOC levels=1,2 markdown formatted bullet hierarchy]: # "Table of content"\n\n# Table of content\n- [Installation](#installation)\n  - [Optional dependencies](#optional-dependencies)\n- [Usage](#usage)\n  - [Auto-completion](#auto-completion)\n- [Development](#development)\n  - [Minimal command](#minimal-command)\n\n# Installation\n\n- pip: `pip install delfino`\n- Poetry: `poetry add -D delfino`\n- Pipenv: `pipenv install -d delfino`\n\nor \n\n- pip: `pip install delfino[completion]`\n- Poetry: `poetry add -D delfino[completion]`\n- Pipenv: `pipenv install -d delfino[completion]`\n\nto enable [auto-completion](#auto-completion).\n\n## Optional dependencies\n\nEach project may use different sub-set of commands. Therefore, dependencies of all commands are optional and checked only when the command is executed.\n\nUsing `[all]` installs all the [optional dependencies](https://setuptools.pypa.io/en/latest/userguide/dependency_management.html#optional-dependencies) used by all the built-in commands. If you want only a sub-set of those dependencies, there are finer-grained groups available:\n\n- For top-level parameters:\n  - `completion` - for `--show-completion` and `--install-completion`\n- For individual commands (matches the command names):\n  - `upload_to_pypi`\n  - `build_docker`\n  - `typecheck`\n  - `format`\n- For groups of commands:\n  - `test` - for testing and coverage commands\n  - `lint` - for all the linting commands\n- For groups of groups:\n  - `verify_all` - same as `[typecheck,format,test,lint]`\n  - `all` - all optional packages\n\n## Configuration\n\nDelfino will assume certain project structure. However, you can customize it to match your own by overriding the default values in the `pyproject.toml` file. Here are the defaults that you can modify:\n\n```toml\n[tool.delfino]\nreports_directory = "reports"\nsources_directory = "src"\ntests_directory = "tests"\nlocal_commands_directory = "commands"\ntest_types = ["unit", "integration"]\nverify_commands = ["format", "lint", "typecheck", "test-all"]\ndisable_pre_commit = false\n\n[tool.delfino.typecheck]\nstrict_directories = [\'src\']\n\n# By default, all commands will be enabled. Use `enable_commands` or `disable_commands` \n# to show only a subset of commands. If both used, disabled commands are subtracted \n# from the set of enabled commands. \n# [tool.delfino.plugins.<PLUGIN_NAME_A>]\n# enable_commands = [<COMMAND_NAME>]\n# disable_commands = [<COMMAND_NAME>]\n# [tool.delfino.plugins.<PLUGIN_NAME_B>]\n# enable_commands = [<COMMAND_NAME>]\n# disable_commands = [<COMMAND_NAME>]\n\n# You can pass additional arguments to commands per project\n# See https://github.com/radeklat/delfino/blob/main/src/delfino/commands/pass_args.py for usage.\n# [tool.delfino.plugins.<PLUGIN>.<COMMAND>]\n# pass_args = \'--capture=no\'\n\n[tool.delfino.dockerhub]\nusername = ""\nbuild_for_platforms = [\n    "linux/amd64",\n    "linux/arm64",\n    "linux/arm/v7",\n]\n```\n\n# Usage\n\nRun `delfino --help` to see all available commands and their usage.\n\n## Auto-completion\n\nYou can either attempt to install completions automatically with:\n\n```shell script\ndelfino --install-completion\n```\n\nor generate it with:\n\n```shell script\ndelfino --show-completion\n```\n\nand manually put it in the relevant RC file.\n\nThe auto-completion implementation is dynamic so that every time it is invoked, it uses the current project. Each project can have different commands or disable certain commands it doesn\'t use. And dynamic auto-completion makes sure only the currently available commands will be suggested.\n\nThe downside of this approach is that evaluating what is available each time is slower than a static list of commands.\n\n# Development\n\nDelfino is a simple wrapper around [Click](https://click.palletsprojects.com). It allows you to add custom, project-specific [commands](https://click.palletsprojects.com/en/8.0.x/quickstart/#basic-concepts-creating-a-command). Let\'s call them Delfino commands or just commands. Commands are expected in the root of the project, in a Python package called `commands`. Any sub-class of [`click.Command`](https://click.palletsprojects.com/en/8.0.x/api/#click.Command) in any `.py` file in this folder will be automatically used by Delfino.\n\n## Minimal command\n\n<!-- TODO(Radek): Delfino expects `pyproject.toml` configured. -->\n<!-- TODO(Radek): Delfino expects Poetry or Pipenv to be available. -->\n\n1. Create the `commands` package:\n   ```shell script\n   mkdir commands\n   touch commands/__init__.py\n   ```\n2. Create a file `commands/command_test.py`, with the following content:\n   ```python\n   import click\n   \n   @click.command()\n   def command_test():\n       """Tests commands placed in the `commands` folder are loaded."""\n       print("✨ This command works! ✨")\n   ```\n3. See if Delfino loads the command. Open a terminal and in the root of the project, call: `delfino --help`. You should see something like this:\n   ```text\n   Usage: delfino [OPTIONS] COMMAND [ARGS]...\n   \n   Options:\n     --help  Show this message and exit.\n   \n   Commands:\n     ...\n     command-test            Tests commands placed in the `commands` folder...\n     ...\n   ```\n4. Run the command with `delfino command-test`\n\n<!--\n## Advanced Command\n\nDelfino adds optional bits of functionality on top of Click. The following example demonstrates some of those:\n\n```python\nimport click\n\nfrom delfino.contexts import pass_app_context, AppContext\nfrom delfino.validation import assert_pip_package_installed, pyproject_toml_key_missing\n\n@click.command()\n# The `pass_app_context` decorator adds `AppContext` as the first parameter.\n@pass_app_context\ndef command_test(app_context: AppContext):\n   """Tests commands placed in the `commands` folder are loaded."""\n   # Test optional dependencies. Any failing assertion will be printed as:\n   # Command \'<NAME>\' is misconfigured. <ASSERTION ERROR MESSAGE> \n   assert_pip_package_installed("delfino")\n   \n   # AppContext contain a parsed `pyproject.toml` file.\n   # Commands can add their config under `[tool.delfino.commands.<COMMAND_NAME>]`.\n   assert "command_test" in app_context.pyproject_toml.tool.delfino.commands, \\\n       pyproject_toml_key_missing("tool.delfino.commands.command_test")\n   \n   print(app_context.pyproject_toml.tool.delfino.commands["command-test"])\n```\n-->\n',
    'author': 'Radek Lát',
    'author_email': 'radek.lat@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/radeklat/delfino',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



import os
import shutil

import click

from delfino.constants import ENTRY_POINT, PackageManager
from delfino.models.app_context import AppContext


def print_no_issues_found():
    click.secho("✔ No issues found.", fg="green")


_HEADER_LEVEL_CHARACTERS = {1: "#", 2: "=", 3: "-"}


def print_header(text: str, level: int = 1, icon: str = ""):
    icon_pad = " " if icon else ""

    padding_character = _HEADER_LEVEL_CHARACTERS[level]
    if os.getenv("CIRCLECI", ""):
        padding_length = 80
    else:
        padding_length = max(shutil.get_terminal_size((80, 20)).columns - (len(icon) * 2), 0)

    padding = f"\n{{:{padding_character}^{padding_length}}}\n"
    if level == 1:
        text = text.upper()
    print(padding.format(f" {icon}{icon_pad}{text} {icon[::-1 if len(icon) > 1 else 1]}{icon_pad}"))


def run_command_example(command: click.Command, app_context: AppContext) -> str:
    prefix = ""
    if app_context.package_manager != PackageManager.UNKNOWN:
        prefix = f"{app_context.package_manager.value} run "
    return f"{prefix}{ENTRY_POINT} {command.name}"



import os
import shlex
import subprocess
from enum import Enum
from logging import getLogger
from typing import Any, Dict, Optional, Tuple

import click

from delfino.backports import shlex_join
from delfino.utils import ArgsType

_LOG = getLogger(__name__)


class OnError(Enum):
    PASS = "pass"
    """Same as ``check=False``. Use when you're handling the return code yourself."""

    EXIT = "exit"
    """Same as ``check=True`` + red print of stdout/stderr + ``raise click.exception.Exit()``.
    Use when running commands where non-zero return code indicates an error, not a failed check."""

    ABORT = "abort"
    """Same as ``check=True`` + print of stdout/stderr + ``raise click.Abort()``.
    Use when running commands where non-zero return code indicates a failed check, not an error."""


def _normalize_args(args: ArgsType, shell: bool) -> Tuple[ArgsType, str]:
    if shell:  # when `shell`, `args` must be a string
        if isinstance(args, str):
            return args, args

        if isinstance(args, bytes):
            args_decoded = args.decode()
            return args_decoded, args_decoded

        args_str = shlex_join(map(str, args))
        return args_str, args_str

    # when not `shell`, `args` must be a `Sequence`
    if isinstance(args, str):
        return shlex.split(args), args

    if isinstance(args, bytes):
        args_decoded = args.decode()
        return shlex.split(args_decoded), args_decoded

    args_list = [str(arg) for arg in args]
    return args_list, shlex_join(args_list)


def _patch_env(
    env_update_path: Optional[Dict[str, Any]] = None, env_update: Optional[Dict[str, Any]] = None
) -> Dict[str, str]:
    modified_env = os.environ.copy()

    if env_update_path or env_update:
        if env_update_path is None:
            env_update_path = {}

        if env_update is None:
            env_update = {}

        for key, value in env_update_path.items():
            modified_env[key] = str(value) + (":" + modified_env[key] if key in modified_env else "")

        modified_env.update({key: str(value) for key, value in env_update.items()})

    return modified_env


def _called_process_error_to_click_exception(
    args: ArgsType, on_error: OnError, exc: subprocess.CalledProcessError
) -> Exception:
    if on_error == OnError.EXIT:
        click.secho(f"\nError ({exc.returncode}) when calling {args!r}:", fg="red")

        if exc.stdout:
            click.secho(exc.stdout.decode(), fg="red")
        if exc.stderr:
            click.secho(exc.stderr.decode(), fg="red")

        return click.exceptions.Exit(code=exc.returncode)

    if exc.stdout:
        print(exc.stdout.decode())
    if exc.stderr:
        print(exc.stderr.decode())

    return click.exceptions.Abort()


def run(
    args: ArgsType,
    *popenargs,
    on_error: OnError,
    env_update_path: Optional[Dict[str, Any]] = None,
    env_update: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> subprocess.CompletedProcess:
    """Modified version of ``subprocess.run``.

    Args:
        args: Program to run with all it's arguments. Passing a sequence is recommended. An
            attempt to parse a string (not bytes) will be made but may lead to an unintended
            split.
        *popenargs: Additional positional arguments passed directly to ``subprocess.run``.
        on_error: Similar to ``Popen(check)`` but differentiating between ``click.exceptions.Exit``
            and ``click.Abort``.
        env_update_path: A dict of path-like environment variables to update. If this variable already
            exists, the value will be pre-pended with a ":".
        env_update: Similar to ``env_update_path`` but any existing variables are replaced.
        **kwargs: Additional keyword arguments passed directly to ``subprocess.run``.
    """
    args, printable_args = _normalize_args(args, kwargs.get("shell", False))
    kwargs["env"] = _patch_env(env_update_path, env_update)

    _LOG.debug(printable_args)

    try:
        return subprocess.run(args, *popenargs, check=on_error != OnError.PASS, **kwargs)
    except subprocess.CalledProcessError as exc:
        raise _called_process_error_to_click_exception(args, on_error, exc) from exc



from functools import update_wrapper
from typing import Any, Callable, TypeVar, cast
from warnings import warn

import click

_Func = TypeVar("_Func", bound=Callable[..., Any])


def filepaths_argument(func: _Func) -> _Func:
    warn(
        f"`@filepaths_argument` from `{__name__}` it deprecated. Use `decorators.files_folders_option` instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    @click.argument("filepaths", nargs=-1, type=click.Path(exists=True))
    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        return ctx.invoke(func, *args, **kwargs)

    return update_wrapper(cast(_Func, new_func), func)
